/*
 *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
 *	This will be free software, but only when it is finished.
 */

#define	RFC974		/* If BIND, check that TCP SMTP service is enabled */

#define	TIMEOUT		(10*60)		/* timeout in seconds, per exchange */

#include <stdio.h>
#include "hostenv.h"
#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <sysexits.h>
/* #include <strings.h> */ /* poorly portable.. */
#include <varargs.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/param.h>
#include <sys/stat.h>

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <mail.h>
#include "ta.h"
#include "malloc.h"


#if	defined(TRY_AGAIN) && defined(USE_RESOLV)
#define	BIND		/* Want BIND (named) nameserver support enabled */
#endif	/* TRY_AGAIN */
#ifdef	BIND
#undef NOERROR				/* pinhead SunOS 5.1 */
#include <arpa/nameser.h>
#include <resolv.h>

#ifndef	BIND_VER
#ifdef	GETLONG
/* 4.7.3 introduced the {GET,PUT}{LONG,SHORT} macros in nameser.h */
#define	BIND_VER	473
#else	/* !GETLONG */
#define	BIND_VER	472
#endif	/* GETLONG */
#endif	/* !BIND_VER */
#endif	/* BIND */

#if	defined(BIND_VER) && (BIND_VER >= 473)
typedef u_char msgdata;
#else	/* !defined(BIND_VER) || (BIND_VER < 473) */
typedef char msgdata;
#endif	/* defined(BIND_VER) && (BIND_VER >= 473) */


/* Define all those things which exist on newer BINDs, and which may
   get returned to us, when we make a query with  T_ANY ... */

#ifndef	T_TXT
# define T_TXT 16
#endif
#ifndef T_RP
# define T_RP
#endif
#ifndef T_AFSDB
# define T_AFSDB 18
#endif
#ifndef T_NSAP
# define T_NSAP 22
#endif
#ifndef T_NSAP_PTR
# define T_NSAP_PTR 23
#endif
#ifndef	T_UINFO		/* Defined on several newer BINDs.. */
# define T_UINFO 100
#endif
#ifndef T_UID
# define T_UID 101
#endif
#ifndef T_GID
# define T_GID 102
#endif
#ifndef T_UNSPEC
# define T_UNSPEC 103
#endif
#ifndef T_SA
# define T_SA 200
#endif



#ifndef	L_SET
#define	L_SET	0
#endif	/* L_SET */

#ifndef	IPPORT_SMTP
#define	IPPORT_SMTP	25
#endif 	/* IPPORT_SMTP */

#define	PROGNAME	"smtpclient"	/* for logging */
#define	CHANNEL		"smtp"	/* the default channel name we deliver for */

/* SV_INTERRUPT is defined in <sys/signal.h> on SunOS */
#ifdef	SV_INTERRUPT
#define	SIGNAL_SYSCALLINTERRUPT(X,Y)	\
	{	struct sigvec sv; \
		sigvec(X, (struct sigvec *)NULL, &sv); \
		sv.sv_handler = Y; \
		sv.sv_flags |= SV_INTERRUPT; \
		sigvec(X, &sv, (struct sigvec *)NULL); \
	}
#else	/* !SV_INTERRUPT */
#define	SIGNAL_SYSCALLINTERRUPT(X,Y)	signal(X,Y)
#endif	/* SV_INTERRUPT */

#ifndef	MAXHOSTNAMELEN
#define	MAXHOSTNAMELEN 64
#endif	/* MAXHOSTNAMELEN */

#define MAXFORWARDERS	128	/* Max number of MX rr's that can be listed */

char myhostname[MAXHOSTNAMELEN+1];
char remotehost[MAXHOSTNAMELEN+1];
char remotemsg[2*BUFSIZ];
char ipaddress[30];
char *progname;
char *cmdline, *eocmdline, *logfile, *msgfile;
char *myaddrs = NULL;
int nmyaddrs = 0;
int pid;
int debug = 0;
int conndebug = 0;
#ifdef	TIMEOUT
int timeout = 0;		/* how long do we wait for response? (sec.) */
int gotalarm =0;		/* indicate that alarm happened! */
#endif
int readalready = 0;		/* does buffer contain valid message data? */
int wantreserved = 0;		/* open connection on secure (reserved) port */
int statusreport = 0;		/* put status reports on the command line */
int noMX = 0;			/* do not use MX records to deliver mail */
int firstmx = 0;		/* [Matti Aarnio] error in smtpwrite("HELO".. */
int mxcount = 0;
int checkwks = 0;
FILE *logfp;

int D_alloc = 0;		/* Memory usage debug */

#ifdef	lint
#undef	putc
#define	putc	fputc
#endif	/* lint */

/* Extended SMTP flags -- can downgrade from 8-bit to 7-bit while at transport
   IF  MIME-Version: is present, AND Content-Transfer-Encoding: 8BIT */

/* Following options can be declared in ESMTP  EHLO responce  */
#define ESMTP_SIZEOPT  0x0001
#define ESMTP_8BITMIME 0x0002
#define ESMTP_DSN      0x0004

int  ehlo_capabilities = 0;
long ehlo_sizeval      = 0;
char esmtp_on_banner = -1;


extern int errno;
extern void warning(), report();
#ifndef MALLOC_TRACE
extern univptr_t emalloc();
#endif
extern char *strerror();
extern char *strchr();
extern char *strrchr();
extern char *dottedquad();
extern char *optarg;
extern int optind;
extern int deliver();
extern int writebuf __((FILE *fp, char *buf, int len, int downgrade));
extern int appendlet __((FILE *fp, FILE *verboselog, int mfp, int hsize, int msize, int downgrade));
extern int smtpopen __((char *host, FILE *fparr[], FILE *verboselog));
extern int smtpconn __((char *host, FILE *fparr[], FILE *verboselog));
extern int smtp_ehlo __((FILE *smtpfp[], FILE *verboselog, char *fmt, char *arg));
extern int ehlo_check __((char *buf));
/* XX: should use  <stdarg.h> -- on machines where it is available.. */
/* extern int smtpwrite __((FILE *smtpfp[2], FILE *verboselog, int saverpt, char *fmt, ... )); */
extern int smtpwrite();
extern int emptyline(), process();
extern void diagnostics();
extern void notaryreport __((char *rcpt, char *acct, char *status));
extern void notarystatsave __((char *smtpstatline));
extern char *notaryacct __((int rc, char *ok_string));

extern int  cte_check __((struct rcpt *rp));
extern void downgrade_headers __((struct rcpt *rp, int downgrade, FILE *verboselog));
extern void check_7bit_downgrade __((int mailfd, int *flag_ptr));

extern void prversion();
extern int makeconn();
extern int vcsetup();
extern void hp_init();
extern char **hp_getaddr(), **hp_nextaddr();
extern void hp_setalist();
#ifdef	BIND
extern int rightmx();
extern int h_errno;
extern int res_mkquery(), res_send(), dn_skipname(), dn_expand();
# ifdef RFC974
extern int getrr(), getmxrr();
extern int getrrtype();
# endif /* RFC974 */
#endif	/* BIND */
extern int matchroutermxes();
extern SIGNAL_TYPE sig_pipe();
#ifdef	TIMEOUT
extern SIGNAL_TYPE sig_alarm();
#endif	/* TIMEOUT */
extern int getmyhostname();
extern void stashmyaddresses();
extern int cistrcmp(), cistrncmp();


FILE *verboselog = NULL;


int
main(argc, argv)
	int argc;
	char *argv[];
{
	char file[MAXPATHLEN+1];
	char *channel, *host;
	int i, fd, errflg, c, smtpstatus;
	FILE *smtpfp[2];
	struct ctldesc *dp;
#ifdef	BIND
	int checkmx = 0;	/* check all destination hosts for MXness */
#endif	/* BIND */

	pid = getpid();
	ipaddress[0] = '\0';
	msgfile = "?";
	getout = 0;
	cmdline = &argv[0][0];
	eocmdline = cmdline;
	for (i = 0; i < argc; ++i)
	  eocmdline += strlen(argv[i]) + 1;
	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
	  signal(SIGINT, wantout);
	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
	  signal(SIGHUP, wantout);
	if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
	  signal(SIGTERM, wantout);
#if 0
	SIGNAL_SYSCALLINTERRUPT(SIGPIPE, sig_pipe);
#else
	signal(SIGPIPE, SIG_IGN);
#endif
#ifdef	TIMEOUT
	SIGNAL_SYSCALLINTERRUPT(SIGALRM, sig_alarm);
	timeout = TIMEOUT;
#endif	/* TIMEOUT */

	progname = PROGNAME;
	errflg = 0;
	channel = CHANNEL;
	wantreserved = debug = statusreport = 0;
	logfile = NULL;
	myhostname[0] = '\0';
	remotemsg[0] = '\0';
	remotehost[0] = '\0';
	while ((c = getopt(argc, argv, "c:deh:l:rsxDE:T:VW")) != EOF) {
	  switch (c) {
	  case 'c':		/* specify channel scanned for */
	    channel = emalloc(strlen(optarg)+1);
	    strcpy(channel, optarg);
	    break;
	  case 'd':		/* turn on debugging output */
	    debug = 1;
	    break;
	  case 'e':		/* expensive MX checking for *all* addresses */
#ifdef	BIND
	    checkmx = 1;
#else  /* !BIND */
	    ++errflg;
	    fprintf(stderr,
		    "%s: -e unavailable, no nameserver support!\n",
		    progname);
#endif /* BIND */
	    break;
	  case 'h':		/* my hostname */
	    strcpy(myhostname, optarg);
	    break;
	  case 'l':		/* log file */
	    logfile = emalloc(strlen(optarg)+1);
	    strcpy(logfile, optarg);
	    break;
	  case 'r':		/* use reserved port for SMTP connection */
	    wantreserved = 1;
	    break;
	  case 's':		/* report status to command line */
	    statusreport = 1;
	    break;
	  case 'x':		/* don't use MX records lookups */
	    noMX = 1;
	    break;
	  case 'D':		/* only try connecting to remote host */
	    conndebug = 1;
	    break;
	  case 'E':		/* don't do EHLO, unless target system
				   has "ESMTP" on its banner */
	    esmtp_on_banner = 0;
	  case 'T':		/* specify Timeout in seconds */
#ifdef	TIMEOUT
	    if ((timeout = atoi(optarg)) < 5) {
	      fprintf(stderr,
		      "%s: illegal timeout: %d\n",
		      argv[0], timeout);
	      ++errflg;
	    }
#else
	    fprintf(stderr,
		    "%s: Timeout not enabled in code!\n",
		    argv[0]);
#endif
	    break;
	  case 'V':
	    prversion("smtp");
	    exit(0);
	    break;
	  case 'W':		/* Enable RFC974 WKS checks */
	    checkwks = 1;
	    break;
	  default:
	    ++errflg;
	    break;
	  }
	}
	if (errflg || optind != argc - 1) {
	  fprintf(stderr,
		  "Usage: %s [-c channel] host\n", argv[0]);
	  exit(EX_USAGE);
	}
	host = emalloc(strlen(argv[optind]) + 1);
	strcpy(host, argv[optind]);

	if (myhostname[0] == 0) {
	  /* Default it only when not having an explicite value
	     in it..   James S MacKinnon <jmack@Phys.UAlberta.Ca> */
	  getmyhostname(myhostname, sizeof myhostname);
	  stashmyaddresses(myhostname);
	}

	if (conndebug && !debug) {
	  firstmx = 0;
	  smtpconn(host, smtpfp, verboselog);
	  exit(0);
	}

	if (logfile != NULL) {
	  if ((fd = open(logfile, O_CREAT|O_APPEND|O_WRONLY, 0644)) < 0)
	    fprintf(stderr,
		    "%s: cannot open logfile \"%s\"!\n",
		    argv[0], logfile);
	  else
	    logfp = (FILE *)fdopen(fd, "a");
	} else
		logfp = NULL;

	/* defer opening a connection until we know there is work */
	smtpstatus = EX_DATAERR;
	while (!getout && fgets(file, sizeof file, stdin)) {
	  if (emptyline(file, sizeof file))
	    break;
#ifdef	BIND
	  if (checkmx)
	    dp = ctlopen(file, channel, host, &getout, rightmx, matchroutermxes);
	  else
#endif /* BIND */
	    dp = ctlopen(file, channel, host, &getout, NULL, matchroutermxes);
	  if (dp == NULL)
	    continue;
	  if (verboselog) {
	    fclose(verboselog);
	    verboselog = NULL;
	  }
	  if (dp->verbose) {
	    verboselog = (FILE *)fopen(dp->verbose,"a");
	    if (verboselog)
	      setbuf(verboselog,NULL);
	  }
	  if (smtpstatus == EX_DATAERR)
	    smtpstatus = smtpopen(host, smtpfp, verboselog);
	  smtpstatus = process(dp, smtpstatus, host, smtpfp, verboselog);
	  ctlclose(dp);
	}
	if (smtpstatus == EX_OK) {
	  smtpstatus = smtpwrite(smtpfp, verboselog, 0, "QUIT");
	  fclose(smtpfp[0]);
	  fclose(smtpfp[1]);
	}
	if (verboselog != NULL)
		fclose(verboselog);
	if (logfp != NULL)
		fclose(logfp);
	exit(smtpstatus);
	/* NOTREACHED */
	return smtpstatus;
}

int
process(dp, smtpstatus, host, smtpfp, verboselog)
	struct ctldesc *dp;
	int smtpstatus;
	FILE *smtpfp[], *verboselog;
	char *host;
{
	struct rcpt *rp, *rphead;
	int loggedid;

	readalready = 0; /* ignore any previous message data cache */
	loggedid = 0;

	for (rp = rphead = dp->recipients; rp != NULL; rp = rp->next) {
	  if (rp->next == NULL
	      || rp->addr->link != rp->next->addr->link
	      || rp->newmsgheader != rp->next->newmsgheader) {
	    if (smtpstatus == EX_OK) {
	      if (logfp != NULL && !loggedid) {
		loggedid = 1;
		fprintf(logfp,
			"%d#\t%s: %s\n",
			pid, dp->msgfile,
			dp->logident);
		fflush(logfp);
	      }
	      smtpstatus = deliver(smtpfp, verboselog, rphead, rp->next,
				   dp->msgfd, dp->msgbodyoffset);
	      if (smtpstatus != EX_OK) {
		fclose(smtpfp[0]);
		fclose(smtpfp[1]);
		smtpstatus = smtpopen(host, smtpfp, verboselog);
	      }
	      rphead = rp->next;
	    } else {
	      while (rphead != rp->next) {
		notaryreport(rp->addr->user,"failed",NULL);
		diagnostic(rphead, smtpstatus, "%s", remotemsg);
		rphead = rphead->next;
	      }
	    }
	  }
	}
	return smtpstatus;
}
/*
 * deliver - deliver the letter in to user's mail box.  Return
 *	     errors and requests for further processing in the structure
 */

int
deliver(smtpfp, verboselog, startrp, endrp, messagefd, msgoffset)
	FILE *smtpfp[2], *verboselog;
	struct rcpt *startrp, *endrp;
	int messagefd;
	long msgoffset;
{
	struct rcpt *rp;
	int r, nrcpt, hsize, size, tout;
	int content_kind, flag_downgrade;
	struct stat stbuf;
	char MAILFROMstr[80];
	char *s;

	if (lseek(messagefd, msgoffset, L_SET) < 0L)
	  warning("Cannot seek to message body! (%m)", (char *)NULL);
	strcpy(MAILFROMstr,"MAIL From:<%s>");
	if (ehlo_capabilities & ESMTP_8BITMIME)
	  strcat(MAILFROMstr, " BODY=8BITMIME");
	s = MAILFROMstr + strlen(MAILFROMstr);
	/* Size estimate is calculated in the  ctlopen()  by
	   adding msg-body size to the largest known header size,
	   though excluding possible header and body rewrites.. */
	if (ehlo_capabilities & ESMTP_SIZEOPT)
	  sprintf(s," SIZE=%d",startrp->desc->msgsizeestimate);

	r = smtpwrite(smtpfp, verboselog, 1, MAILFROMstr,
		      startrp->addr->link->user);
	if (r != EX_OK) {
	  for (rp = startrp; rp != endrp; rp = rp->next) {
	    notaryreport(rp->addr->user,notaryacct(r,"*NOTOK*"),NULL);
	    diagnostic(rp, r, "%s", remotemsg);
	  }
	  smtpwrite(smtpfp, verboselog, 0, "RSET");
	  return r;
	}
	nrcpt = 0;
	for (rp = startrp; rp != endrp; rp = rp->next) {
	  r = smtpwrite(smtpfp, verboselog, 1, "RCPT To:<%s>", rp->addr->user);
	  if (r != EX_OK) {
	    notaryreport(rp->addr->user, notaryacct(r,"*NOTOK"),NULL);
	    diagnostic(rp, r, "%s", remotemsg);
	  } else {
	    ++nrcpt;
	    rp->status = EX_OK;
	  }
	}
	if (nrcpt == 0) {
	  /* all the RCPT To addresses were rejected, so reset server */
	  smtpwrite(smtpfp, verboselog, 0, "RSET");
	  return EX_UNAVAILABLE;
	}
	if ((r = smtpwrite(smtpfp, verboselog, 1, "DATA")) != EX_OK) {
	  for (rp = startrp; rp != endrp; rp = rp->next)
	    if (rp->status == EX_OK) {
	      notaryreport(rp->addr->user,notaryacct(r,"*NOTOK"),NULL);
	      diagnostic(rp, EX_TEMPFAIL, "%s", remotemsg);
	    }
	  smtpwrite(smtpfp, verboselog, 0, "RSET");
	  return r;
	}
	/* Headers are 7-bit stuff -- says MIME specs */


	/* Content-Transfer-Encoding: 8BIT ? */
	content_kind = cte_check(startrp);
	flag_downgrade = ((content_kind == 8) &&
			  ((ehlo_capabilities & ESMTP_8BITMIME) == 0));
	/* If it does not contain MIME headers, and it contains 8-bit
	   chars, it is in trouble!  Mark the checkup with  -1       */
	if (content_kind == 0 && (ehlo_capabilities & ESMTP_8BITMIME) == 0)
	  flag_downgrade = -1;
	if (flag_downgrade) {
	  check_7bit_downgrade(messagefd, &flag_downgrade);
	  if (flag_downgrade & 1) /* Downgrade 'em only if really need it! */
	    downgrade_headers(startrp, flag_downgrade, verboselog);
	}

	if (verboselog) {
	  char **hdrs = *(startrp->newmsgheader);
	  if (*(startrp->newmsgheadercvt) != NULL && (flag_downgrade & 1))
	    hdrs = *(startrp->newmsgheadercvt);
	  fprintf(verboselog,
		  "Downgraded headers:  ContentKind=%d, flag=%d\n------\n",
		  content_kind,flag_downgrade);
	  while (hdrs && *hdrs)
	    fprintf(verboselog,"%s\n",*hdrs++);
	}

	if ((hsize = writeheaders(startrp, smtpfp[1], "\r\n",
				  flag_downgrade & 1, 0)) < 0) {
	  for (rp = startrp; rp != endrp; rp = rp->next)
	    if (rp->status == EX_OK) {
	      notaryreport(rp->addr->user,notaryacct(r,"*NOTOK*"),
			   "466 (Message write timed out)"); /* XX: FIX THE STATUS? */
	      diagnostic(rp, EX_TEMPFAIL, "%s", "header write error");
	    }
	  if (verboselog)
	    fprintf(verboselog,"Writing headers after DATA failed\n");
	  smtpwrite(smtpfp, verboselog, 0, "RSET");
	  return EX_TEMPFAIL;
	}
	
	/* append message body itself */
	if (fstat(messagefd, &stbuf) >= 0)
	  size = hsize + stbuf.st_size - msgoffset;
	else
	  size = -hsize;
	if ((r = appendlet(smtpfp[1], verboselog, messagefd,
			   hsize, size, flag_downgrade)) != EX_OK) {
	  for (rp = startrp; rp != endrp; rp = rp->next)
	    if (rp->status == EX_OK) {
	      notaryreport(rp->addr->user, notaryacct(r,"*NOTOK*"),
			   "466 (Message write timed out)"); /* XX: FIX THE STATUS? */
	      diagnostic(rp, r, "%s", remotemsg);
	    }
	  /*
	   * The failure was due to an I/O error either in the local
	   * message file, because the socket got torn down under our
	   * feet, or because the other end is out of sync.  In case 1
	   * and 3 we want to try to convince the remote to abort the
	   * transaction.  Issuing a RSET is empirically a Good Thing
	   * in this situation, as suggested by Matti Aarnio.  If we
	   * are still in DATA mode the RSET is harmless.  If we are
	   * in command mode, the message is aborted and the following
	   * "." will generate a 500 Command unrecognized which is
	   * also harmless.
	   */
	  smtpwrite(smtpfp, verboselog, 0, "RSET");
	}
	/*
	 * This is the one place where we *have* to wait forever because
	 * there is no reliable way of aborting the transaction.
	 * Note that a good and useful approximation to "forever" is one day.
	 * Murphy's Law you know: Connections will hang even when they can't.
	 */
#ifdef	TIMEOUT
	tout = timeout, timeout = 60*60*24;
#endif
	r = smtpwrite(smtpfp, verboselog, 1, ".");

#ifdef	TIMEOUT
	timeout = tout;
#endif
	if (logfp != NULL) {
	  if (r != EX_OK)
	    fprintf(logfp, "%d#\t%s\n", pid, remotemsg);
	  else
	    fprintf(logfp, "%d#\t%d bytes, %d in header\n",
		    pid, size, hsize);
	  fflush(logfp);
	}
	for (rp = startrp; rp != endrp; rp = rp->next)
	  if (rp->status == EX_OK) {
	    notaryreport(rp->addr->user,
			 notaryacct(r,"relayed"), /* XX: NOTARY DRPT ?? */
			 NULL);
	    diagnostic(rp, r, "%s", remotemsg);
	  }
	if (r != EX_OK)
	  smtpwrite(smtpfp, verboselog, 0, "RSET");
	return r;
}

/*
 * appendlet - append letter to file pointed at by fd
 */

static char let_buffer[BUFSIZ*8];


int
appendlet(fp, verboselog, mfd, hsize, msize, downgrade)
	FILE *fp, *verboselog;
	int mfd, hsize, msize, downgrade;
{
	/* downgrade&1  is non-zero  when MIME-tag Content-Transfer-Encoding:
	   is altered from  8BIT to  QUOTED-PRINTABLE.   When   downgrade&2 is
	   non-zero, alteration is from 8BIT to 7BIT.
	 */

	register int i;
	register int bufferfull;
	int lastwasnl;

	writebuf(fp, (char *)NULL, 0, 0);	/* magic init. */

	/* can we use cache of message body data */
	if (readalready != 0) {
	  if (writebuf(fp, let_buffer, readalready, downgrade) != readalready)
	    return EX_IOERR;
	  if (let_buffer[readalready-1] != '\n')
	    writebuf(fp, "\n", 1, 0);
	  if (statusreport)
	    report("DATA %d (100%%)", readalready+hsize);
	  return EX_OK;
	}

	/* we are assumed to be positioned properly at start of message body */
	bufferfull = 0;
	lastwasnl = 1;	/* we are guaranteed to have a \n after the header */
#ifdef TIMEOUT
	alarm(1*60*60); /* Timeout in 1 hour per let_buffer full... BUFSIZ/h is
			   damn slow.. */
	gotalarm = 0;
#endif
	while ((i = read(mfd, let_buffer, sizeof let_buffer)) != 0) {
	  if (i < 0) {
	    notaryreport(NULL,NULL,"500 (Read error from message file!?)");
	    strcpy(remotemsg, "read error from message file");
	    if (verboselog)
	      fprintf(verboselog,
		      "source read error! (%s)  DATA %d/%d (%d%%)",
		      strerror(errno), hsize, msize, (hsize*100+msize/2)/msize);
	    return EX_IOERR;
	  }
	  lastwasnl = (let_buffer[i-1] == '\n');
	  if (writebuf(fp, let_buffer, i, downgrade) != i) {
#ifdef TIMEOUT
	    if (gotalarm) {
	      sprintf(remotemsg,"500 (let_buffer write timeout!  DATA %d/%d [%d%%])",
		      hsize, msize, (hsize*100+msize/2)/msize);
	      notaryreport(NULL,NULL,remotemsg);
	      if (verboselog)
		fprintf(verboselog,"%s\n",remotemsg);
	      if (bufferfull > 1) readalready = 0;
	      return EX_TEMPFAIL;
	    }
#endif
	    sprintf(remotemsg,
		    "500 (let_buffer write IO-error! [%s] DATA %d/%d [%d%%])",
		    strerror(errno), hsize, msize, (hsize*100+msize/2)/msize);
	    notaryreport(NULL,NULL,remotemsg);
	    if (verboselog)
	      fprintf(verboselog,"%s\n",remotemsg);

	    if (bufferfull > 1) readalready = 0;
	    return EX_IOERR;
	  }
	  if (statusreport) {
	    hsize += i;
	    report("DATA %d/%d (%d%%)",
		   hsize, msize, (hsize*100+msize/2)/msize);
	  }
	  readalready = i;
	  bufferfull++;
	}
	/* we must make sure the last thing we transmit is a CRLF sequence */
	if (!lastwasnl)
	  if (writebuf(fp, "\n", 1, downgrade) != 1) {
#ifdef TIMEOUT
	    if (gotalarm) {
	      sprintf(remotemsg,"500 (let_buffer write timeout!  DATA %d/%d [%d%%])",
		      hsize, msize, (hsize*100+msize/2)/msize);
	      notaryreport(NULL,NULL,remotemsg);
	      if (verboselog)
		fprintf(verboselog,"%s\n",remotemsg);
	      if (bufferfull > 1) readalready = 0;
	      return EX_TEMPFAIL;
	    }
#endif
	    sprintf(remotemsg,
		    "500 (let_buffer write IO-error! [%s] DATA %d/%d [%d%%])",
		    strerror(errno), hsize, msize, (hsize*100+msize/2)/msize);
	    notaryreport(NULL,NULL,remotemsg);
	    if (verboselog)
	      fprintf(verboselog,"%s\n",remotemsg);

	    if (bufferfull > 1) readalready = 0;
	    return EX_IOERR;
	  }

	if (bufferfull > 1)	/* not all in memory, need to reread */
	  readalready = 0;
	return EX_OK;
}

/*
 * Writebuf() is like write(), except all '\n' are converted to "\r\n"
 * (CRLF), and the sequence "\n." is converted to "\r\n..".
 */

int
writebuf(fp, buf, len, downgrade)
	FILE *fp;
	char *buf;
	int len;
	int downgrade;
{
	register char *cp;
	register int n;
	int tlen;
	register int state, column;
	char *i2h = "0123456789ABCDEF";
	static int save = 0;
	static int col  = 0;
	int qp_conv = (downgrade & 1) /* && !(downgrade & 2) */ ;

	if (buf == NULL) {		/* magic initialization */
	  save = 1;
	  col  = 0;
	  return 0;
	}
	state  = save;
	column = col;
	for (cp = buf, n = len, tlen = 0; n > 0; --n, ++cp) {
	  int c = *(unsigned char *)cp;
	  if (qp_conv && column > 70 && c != '\n') {
	    putc('=',fp);
	    putc('\r',fp);
	    putc('\n',fp);
	    /* if (verboselog) fprintf(verboselog,"=\n"); */
	    column = 0;
	    state = 1;
	  }
	  if (qp_conv &&
	      (c == '='  ||  c > 126 ||  (c != '\n' && c < 32))) {

	    putc('=',fp);
	    putc(i2h[(c >> 4) & 15],fp);
	    putc(i2h[(c)      & 15],fp);
	    column += 3;
	    /* if (verboselog) fprintf(verboselog,"=%02X",c); */
	    ++tlen;
	    continue;
	  }

	  if (state && c != '\n') {
	    state = 0;
	    if (c == '.') {
	      if (putc(c, fp) == EOF || putc(c, fp) == EOF) {
		notaryreport(NULL,NULL,"500 (body write error, 1)");
		strcpy(remotemsg, "write error 1");
		return EOF;
	      }
	      ++column;
	      /* if (verboselog) fprintf(verboselog,".."); */
	    } else {
	      if (putc(c, fp) == EOF) {
		notaryreport(NULL,NULL,"500 (body write error, 2)");
		strcpy(remotemsg, "write error 2");
		return EOF;
	      }
	      /* if (verboselog) putc(c, verboselog); */
	    }
	    ++column;
	    ++tlen;
	  } else if (c == '\n') {
	    if (putc('\r', fp) == EOF || putc(c, fp) == EOF) {
	      notaryreport(NULL,NULL,"500 (body write error, 3)");
	      strcpy(remotemsg, "write error 3");
	      return EOF;
	    }
	    /* if (verboselog) putc(c, verboselog); */
	    ++tlen;
	    column = 0;
	    state = 1;
	  } else {
	    if (putc(c, fp) == EOF) {
	      notaryreport(NULL,NULL,"500 (body write error, 4)");
	      strcpy(remotemsg, "write error 4");
	      return EOF;
	    }
	    /* if (verboselog) putc(c, verboselog); */
	    ++column;
	    ++tlen;
	  }
	}
	col  = column;
	save = state;
	return tlen;
}

int
ehlo_check(buf)
char *buf;
{

	if (strncmp(buf,"8BITMIME",8)==0) {
	  ehlo_capabilities |= ESMTP_8BITMIME;
	} else if (strcmp(buf,"XDSN")==0 ||
		   strcmp(buf,"DSN")==0) {
	  ehlo_capabilities |= ESMTP_DSN;
	} else if (strncmp(buf,"SIZE ",5)==0 ||
		   strcmp(buf,"SIZE") == 0) {
	  ehlo_capabilities |= ESMTP_SIZEOPT;
	  ehlo_sizeval = -1;
	  if (buf[5] == ' ')
	    sscanf(buf+5,"%d",&ehlo_sizeval);
	}
	return 0;
}

/* Flag that banner contained "ESMTP" (case insensitive) */
void
esmtp_banner_check(str)
char *str;
{
	char *s = str;
	while (*s) {
	  while (*s && *s != 'e' && *s != 'E') ++s;
	  if (!*s) return;
	  if (cistrncmp(s,"ESMTP",5)==0) {
	    esmtp_on_banner = 1;
	    return;
	  }
	  ++s;
	}
}


int
smtpopen(host, fparr, verboselog)
	char *host;
	FILE *fparr[], *verboselog;
{
	int i;

	if (debug) {
	  fprintf(logfp, "%d#\tsmtpopen: connecting to %s\n", pid, host);
	  fflush(logfp);
	}
	firstmx = 0;
	mxcount = 0;
	do {
	  if ((i = smtpconn(host, fparr, verboselog)) != EX_OK)
	    return i;
	  if (esmtp_on_banner) {
	    /* Either it is not tested, or it is explicitely
	       desired to be tested, and was found! */
	    i = smtp_ehlo(fparr, verboselog, "EHLO %s", myhostname);
	    if (i == EX_TEMPFAIL) {
	      /* There are systems, which hang up on us
		 when we present "EHLO".. Do normal HELO.. */
	      if ((i = smtpconn(host, fparr, verboselog)) != EX_OK)
		return i;
	      i = EX_TEMPFAIL;
	    }
	  }
	  if (i == EX_OK) {
	    if (verboselog)
	      fprintf(verboselog,
		      "  EHLO responce flags = 0x%02x\n",
		      ehlo_capabilities);
	  } else
	    i = smtpwrite(fparr, verboselog, 1, "HELO %s", myhostname);
	} while (i == EX_TEMPFAIL && firstmx <= mxcount);
	if (debug) {
	  fprintf(logfp, "%d#\tsmtpopen: status = %d\n", pid, i);
	  fflush(logfp);
	}
	return i;
}

struct mxdata {
	msgdata	*host;
	u_short	pref;
};

int
smtpconn(host, fparr, verboselog)
	char *host;
	FILE *fparr[], *verboselog;
{
	int	i, r, retval, smtpfd;
	u_long	ipaddr;
	struct hostent *hp, hent;
	char	hbuf[MAXHOSTNAMELEN+1], *addrs[2];
	struct mxdata mxh[MAXFORWARDERS];
	int	mxcnt = 0;

#ifdef	BIND
	h_errno = 0;
#endif	/* BIND */

	if (debug) {
	  fprintf(logfp, "%d#\tsmtpconn: host = %s\n", pid, host);
	  fflush(logfp);
	}
	if (verboselog)
	  fprintf(verboselog,"SMTP: Connecting to host: %s\n",host);
	if (*host == '[') {	/* hostname is IP address domain literal */
	  char *cp, *hcp, buf[BUFSIZ];

	  for (cp = buf, hcp = host + 1;
	       isascii(*hcp) && (isdigit(*hcp) || *hcp == '.');
	       ++cp, ++hcp)
	    *cp = *hcp;
	  *cp = '\0';
	  ipaddr = inet_addr(buf);
	  if (((int)ipaddr) == -1) {
	    sprintf(remotemsg, "500 (bad IP address: %s)", host);
	    notaryreport(NULL,NULL,remotemsg);
	    if (verboselog)
	      fprintf(verboselog,"%s\n", remotemsg);
	    return EX_NOHOST;
	  }
	  hp = &hent;
	  /* don't really care about gethostbyaddr() here */
	  strcpy(hbuf,host);
	  hp->h_name = hbuf;
	  hp->h_addrtype = AF_INET;
	  hp->h_aliases = NULL;
	  hp->h_length = sizeof ipaddr;
	  addrs[0] = (char *)&ipaddr;
	  addrs[1] = NULL;
	  hp_setalist(hp, addrs);
	  retval = makeconn(hp, &smtpfd, verboselog, fparr, mxcnt);
	} else {
	  hbuf[0] = '\0';
	  errno = 0;
#if 0  /* ---------- rip off this gethostbyname() sequence -------- */
	  if ((hp = gethostbyname(host)) != NULL) {
	    strcpy(hbuf,hp->h_name); /* Get canonic form of target name */
	    host = hbuf;	/* Save it, we reuse gethostbyname() */
	  } else {		/* Failed somehow, blah! */
	    if (errno == ETIMEDOUT
#ifdef	XBIND
		|| h_errno == TRY_AGAIN
#endif /* BIND */
		) {
	      sprintf(remotemsg,"466 (gethostbyname(%s): try later)",host);
	      notaryreport(NULL,NULL,remotemsg);
	      if (verboselog)
		fprintf(verboselog,"%s\n",remotemsg);
	      return EX_TEMPFAIL;
	    }
#ifdef	BIND
	    /*
	     * Find the canonical name
	     * (needed when target of a CNAME has no other RR's)
	     */
	    strcpy(hbuf, host);
	    if ((i = getrr(hbuf, sizeof hbuf, (u_short)T_CNAME),2) == 1) {
	      host = hbuf;
	    } else if (h_errno == TRY_AGAIN) {
	      sprintf(remotemsg, "466 (Canonicalize, try later: %s)", host);
	      notaryreport(NULL,NULL,remotemsg);
	      if (verboselog)
		fprintf(verboselog,"%s\n",remotemsg);
	      return EX_TEMPFAIL;
	    } /* else [Matti Aarnio says:]
		 it is possible that MXs or As are available
		 even though CNAME lookups fail (see .UK) */
#else  /* !BIND */
	    sprintf(remotemsg, "466 (canonicalize lookup failed on %s)", host);
	    notaryreport(NULL,NULL,remotemsg);
	    if (verboselog)
	      fprintf(verboselog,"%s\n", remotemsg);
	    /* temporary failure since the router knew it */
	    return EX_TEMPFAIL;	/* was EX_NOHOST */
#endif /* BIND */
	  }
#endif /* ---------- rip off this gethostbyname() sequence -------- */

	  mxh[0].host = NULL;	/* use the hostent we got */
#ifdef	BIND
	  /*
	   * Look for MX RR's. If none found, use the hostentry in hp.
	   * Otherwise loop through all the mxhosts doing gethostbyname's.
	   */
	  mxcnt = 0;
	  if (!noMX) {
	    int rc;
	    if (verboselog)
	      fprintf(verboselog," getmxrr(%s)",host);
	    rc = getmxrr(host, mxh, sizeof mxh/sizeof mxh[0],0,&mxcnt);
	    if (verboselog)
	      fprintf(verboselog," rc=%d, mxh[0].host=%s (host=%s)\n",
		      rc,mxh[0].host,host);
	    hp = NULL;		/* Ruined HP within getmxrr() */
	    switch (rc) {
	      /* remotemsg is generated within getmxrr */
	    case EX_SOFTWARE:
	    case EX_TEMPFAIL:
	    case EX_UNAVAILABLE:
	      return EX_TEMPFAIL;
	    case EX_NOHOST:
	      return EX_NOHOST;
	    case EX_NOPERM:
	      return EX_NOPERM;
	    }
	  }
#endif /* BIND */
	  if (noMX || mxh[0].host == NULL) {
	    errno = 0;
	    /* Either forbidden MX usage, or does not have MX entries! */
	    if (hp == NULL && (hp = gethostbyname(host)) == NULL) {
	      if (errno == ETIMEDOUT
#ifdef	BIND
		  || h_errno == TRY_AGAIN
#endif /* BIND */
		  ) {
		sprintf(remotemsg,"466 (gethostbyname<%s>: try later)",host);
		notaryreport(NULL,NULL,remotemsg);
		if (verboselog)
		  fprintf(verboselog,"%s\n",remotemsg);
		return EX_TEMPFAIL;
	      }

	      if (noMX)
		sprintf(remotemsg,
			"500 (configuration inconsistency. MX usage forbidden, no address in DNS: '%s', errno=%s)",
			host,strerror(errno));
	      else
		sprintf(remotemsg,
			"500 (nameserver data inconsistency. No MX, no address: '%s', errno=%s)",
			host,strerror(errno));
	      notaryreport(NULL,NULL,remotemsg);
	      if (verboselog)
		fprintf(verboselog,"%s\n",remotemsg);
	      /* it was: EX_UNAVAILABLE, but such blocks retrying, thus
		 current EX_TEMPFAIL, which will cause timeout latter on.. */
	      return EX_TEMPFAIL;
	    }
	    retval = makeconn(hp, &smtpfd, verboselog, fparr, mxcnt);
	  } else {
	    retval = EX_UNAVAILABLE;
	    for (i = 0; mxh[i].host != NULL; ++i) {
	      if ((hp = gethostbyname(mxh[i].host)) == NULL) {
		free(mxh[i].host);
		mxh[i].host = NULL;
		firstmx = i+1;
		continue;
	      }
	      r = makeconn(hp, &smtpfd, verboselog, fparr, mxcnt);
	      if (r == EX_OK) {
		while (mxh[i].host != NULL) {
		  free(mxh[i].host);
		  mxh[i++].host = NULL;
		}
		retval = EX_OK;
		break;
	      } else if (r == EX_TEMPFAIL)
		retval = EX_TEMPFAIL;
	      free(mxh[i].host);
	      mxh[i].host = NULL;
	      firstmx = i+1;
	    }
	  }
	}
	if (retval == EX_OK) {
	  strcpy(remotehost, hp->h_name);
	} else
		strcpy(remotehost, host);
	if (debug) {
	  fprintf(logfp,
		  "%d#\tsmtpconn: retval = %d\n", pid, retval);
	  fflush(logfp);
	}
	return retval;
}

int
makeconn(hp, fdp, verboselog, fparr, mxcnt)
	struct hostent *hp;
	int *fdp;
	FILE *fparr[];
	FILE *verboselog;
	int mxcnt;
{
	int retval;

#ifdef	BIND
#ifdef	RFC974
	char hostbuf[MAXHOSTNAMELEN+1];

	if (checkwks && verboselog)
	  fprintf(verboselog,"  makeconn(): checkwks of host %s\n",hp->h_name);

	strcpy(hostbuf, hp->h_name);
	if (checkwks &&
	    getrr(hostbuf, sizeof hostbuf, (u_short)T_WKS, 2) != 1) {
	  sprintf(remotemsg,"550 (WKS checks: no SMTP capability for host %s)",
		  hostbuf);
	  notaryreport(NULL,NULL,remotemsg);
	  if (verboselog)
	    fprintf(verboselog,"%s\n",remotemsg);
	  return EX_UNAVAILABLE;
	}
#endif	/* RFC974 */
#endif	/* BIND */
	
	retval = EX_UNAVAILABLE;
	for (hp_init(hp); *hp_getaddr() != NULL; hp_nextaddr()) {
	  if (matchmyaddress(*hp_getaddr(),hp->h_length)) {
	    notaryreport(NULL,NULL,"500 (Trying to talk with myself!)");
	    sprintf(remotemsg,"Trying to talk with myself!");
	    break;		/* TEMPFAIL or UNAVAILABLE.. */
	  }
	  switch (vcsetup(*hp_getaddr(), hp->h_length, fdp, hostbuf, verboselog)) {
	  case EX_OK:
	    fparr[0] = (FILE*) fdopen(*fdp, "r");
	    fparr[1] = (FILE*) fdopen(*fdp, "w");
	    if (esmtp_on_banner > 0)
	      esmtp_on_banner = 0;
	    retval = smtpwrite(fparr, verboselog, 1, NULL);
	    if (retval != EX_OK)
	      /*
	       * If you want to continue with the next host,
	       * the below should be 'return EX_TEMPFAIL'.
	       */
	      break;		/* try another host address */
	    return EX_OK;
	  case EX_TEMPFAIL:
	    retval = EX_TEMPFAIL;
	    break;
	  }
	}
	return retval;
}

int
vcsetup(ipaddrcp, ipaddrlen, fdp, hostname, verboselog)
	char *ipaddrcp;
	int ipaddrlen, *fdp;
	char *hostname;
	FILE *verboselog;
{
	register int s;
	struct sockaddr_in sad;
	u_short p;
	int errnosave;
	char *se;

	strcpy(ipaddress, dottedquad((struct in_addr *)ipaddrcp));
	if (conndebug)
	  fprintf(stderr, "Trying %s [%s] ... ", hostname, ipaddress);
	/* try grabbing a port */
	sad.sin_family = AF_INET;
	sad.sin_addr.s_addr = 0;
	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	  se = strerror(errno);
	  sprintf(remotemsg, "500 (Internal error, socket: %s)", se);
	  notaryreport(NULL,NULL,remotemsg);
	  if (conndebug)
	    fprintf(stderr, "%s\n", remotemsg);
	  if (verboselog)
	    fprintf(verboselog,"%s\n",remotemsg);
	  return EX_TEMPFAIL;
	}
	if (wantreserved && getuid() == 0) {
	  for (p = IPPORT_RESERVED-1; p >= IPPORT_RESERVED/2; --p) {
	    sad.sin_port = htons(p);
	    if (bind(s, (struct sockaddr *)&sad, sizeof sad) >= 0)
	      break;
	    if (errno != EADDRINUSE && errno != EADDRNOTAVAIL) {
	      char *s = strerror(errno);
	      sprintf(remotemsg, "500 (Internal error, bind: %s)", s);
	      notaryreport(NULL,NULL,remotemsg);
	      if (verboselog)
		fprintf(verboselog,"%s\n", remotemsg);
	      if (conndebug)
		fprintf(stderr, "%s\n", remotemsg);
	      return EX_UNAVAILABLE;
	    }
	  }
	  if (p < IPPORT_RESERVED/2) {
	    sprintf(remotemsg, "too many busy ports");
	    notaryreport(NULL,NULL,"500 (Internal error, too many busy ports)");
	    if (conndebug)
	      fprintf(stderr, "%s\n", remotemsg);
	    if (verboselog)
	      fprintf(verboselog,"%s\n",remotemsg);
	    return EX_TEMPFAIL;
	  }
	}
	bcopy(ipaddrcp, (char *)&sad.sin_addr, ipaddrlen);
	sad.sin_port = htons((u_short)IPPORT_SMTP);
	if (connect(s, (struct sockaddr *)&sad, sizeof sad) >= 0) {
	  int on = 1;
	  *fdp = s;
#if	BSD >= 43
	  setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof on);
#else  /* BSD < 43 */
	  setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, 0, 0);
#endif /* BSD >= 43 */
	  if (conndebug)
	    fprintf(stderr, "connected!\n");
	  return EX_OK;
	}

	errnosave = errno;
	se = strerror(errno);
	sprintf(remotemsg, "500 (connect to %s [%s]: %s)",
		hostname, ipaddress, se);
	notaryreport(NULL,NULL,remotemsg);

	if (conndebug)
		fprintf(stderr, "%s\n", remotemsg);
	if (verboselog)
		fprintf(verboselog,"%s\n",remotemsg);
	switch (errnosave) {	/* from sendmail... */
	case EISCONN:
	case ETIMEDOUT:
	case EINPROGRESS:
	case EALREADY:
	case EADDRINUSE:
	case EHOSTDOWN:
	case ENETDOWN:
	case ENETRESET:
	case ENOBUFS:
	case ECONNREFUSED:
	case ECONNRESET:
	case EHOSTUNREACH:
	case ENETUNREACH:
	case EPERM:
	/* wonder how Sendmail missed this one... */
	case EINTR:
		close(s);
		return EX_TEMPFAIL;
	}
	close(s);
	return EX_UNAVAILABLE;
}

SIGNAL_TYPE
sig_pipe()
{
	if (logfp != NULL) {
	  fprintf(logfp, "%d#\t*** Received SIGPIPE!\n", pid);
	  fflush(logfp);
	  abort();
	}
}

#ifdef	TIMEOUT
SIGNAL_TYPE
sig_alarm()
{
	gotalarm = 1;
}
#endif	/* TIMEOUT */

void
rmsgappend(va_alist)
	va_dcl
{
	va_list ap;
	char *s, *arg;
	char *cp = remotemsg + strlen(remotemsg);
	char *cpend = remotemsg + sizeof remotemsg -1;

	va_start(ap);
	s   = va_arg(ap, char *);

	if (!s) s="(NULL)";
	for (; *s != 0; ++s) {
	  if (*s == '%' && *++s == 's') {
	    arg = va_arg(ap, char *);
	    while (*arg && cp < cpend)
	      *cp++ = *arg++;
	  } else
	    if (cp < cpend)
	      *cp++ = *s;
	}
	*cp = 0;
	va_end(ap);
}

int dflag = 0;

int within_ehlo = 0;

/* VARARGS */
/*int
smtpwrite(smtpfp, verboselog, saverpt, fmt, arg)
	FILE *smtpfp[2], *verboselog;
	int saverpt;
	char *fmt, *arg;		*/
int
smtpwrite(va_alist)
	va_dcl
{
	va_list ap;
	FILE **smtpfp;
	int saverpt;
	FILE *verboselog;
	char *fmt;
	int argc = 0;
	unsigned long argv[20];
	register char *s, *cp;
	int i, response, r, infd, outfd;
	char ch, buf[8096];	/* XX: growl */
	char *se;

	&smtpfp; /* Just make sure it stays in memory .. */

	va_start(ap);
	smtpfp     = va_arg(ap, FILE **);
	verboselog = va_arg(ap, FILE *);
	saverpt    = va_arg(ap, int);
	fmt	   = va_arg(ap, char *);

#ifdef	TIMEOUT
	alarm(timeout);	/* This much total to write and get answer */
	gotalarm = 0;
	/* we don't need to turn alarms off again (yay!) */
#endif	/* TIMEOUT */
	fflush(smtpfp[0]);
	fflush(smtpfp[1]);
	infd = fileno(smtpfp[0]);
	outfd = fileno(smtpfp[1]);

	if (fmt == NULL)
	  va_end(ap);
	else {
	  /* we use raw I/O so we have control over what happens on
	     interrupts, etc. -- I don't like setjmp/longjmp */
	  for (s = fmt, cp = buf; *s != '\0'; ++s) {
	    if (*s == '%') {
	      switch (*++s) {
	      case 's':
		argv[argc] = (unsigned long)va_arg(ap, char*);
		strcpy(cp, (char*)argv[argc]);
		++argc;
		cp += strlen(cp);
		break;
	      case 'd':
		argv[argc] = (unsigned long)va_arg(ap, int);
		sprintf(cp, "%d", (int)argv[argc]);
		++argc;
		cp += strlen(cp);
		break;
	      default:
		*cp++ = '%';
	      case '%':
		*cp++ = *s;
		break;
	      }
	    } else 
	      *cp++ = *s;
	  }
	  *cp++ = '\r';
	  *cp++ = '\n';
	  va_end(ap);
	  if (verboselog)
	    fwrite(buf, cp - buf, 1, verboselog);

	  if ((r = write(outfd, buf, cp - buf)) == -1) {
#ifdef	TIMEOUT
	    if (gotalarm) {
	      strcpy(remotemsg, "Timeout on cmd write");
	      notaryreport(NULL,NULL,"500 (timeout on cmd write)");
	    } else
#endif /* TIMEOUT */
	      {
		se = strerror(errno);
		sprintf(remotemsg, "500 (write to server error: %s)", se);
		notaryreport(NULL,NULL,remotemsg);
	      }
	    if (verboselog)
	      fprintf(verboselog,"%s\n",remotemsg);
	    return EX_TEMPFAIL;
	  } else if (r != cp - buf) {
	    sprintf(remotemsg, "500 (SMTP cmd write failure: Only wrote %d of %d bytes!)", r, cp - buf);
	    notaryreport(NULL,NULL,remotemsg);
	    if (verboselog)
	      fprintf(verboselog,"%s\n",remotemsg);
	    return EX_TEMPFAIL;
	  }
	  if (logfp != NULL) {
	    if (dflag) abort();
	    fprintf(logfp, "%dw\t", pid);
	    fprintf(logfp, fmt,
		    argv[0],argv[1],argv[2],
		    argv[3],argv[4],argv[5]);
	    fprintf(logfp, "\n");
	    fflush(logfp);
	    dflag = 1;
	  }
	}
	i = 2;	/* state variable, beginning of new line */

	if (fmt) {
	  if (strncmp(fmt,"RSET",4) != 0) /* If RSET, append to previous! */
	    *remotemsg = 0;
	  rmsgappend("\r<<- ");
	  rmsgappend(fmt,
		     argv[0],argv[1],argv[2],
		     argv[3],argv[4],argv[5]);
	} else {
	  strcpy(remotemsg,"\r<<- (null)");
	}

	if (debug) {
	  fprintf(logfp, "%d#\tAttempting to read reply\n",pid);
	  fflush(logfp);
	}
	if (statusreport && fmt != NULL) {
	  report(fmt,
		 argv[0],argv[1],argv[2],
		 argv[3],argv[4],argv[5]);
	}
	cp = buf;
	do {
	  r = read(infd, cp, sizeof buf - (cp - buf));
	  if (r > 0) {
	    if (verboselog)
	      fwrite(cp,r,1,verboselog);
	    for (s = cp, cp += r; s < cp; ++s) {
	      switch (i) {
	      	/* i == 0 means we're on last line */
	      case 1:		/* looking for \n */
		if (*s != '\n')
		  break;
		*s = '\0';
		if (within_ehlo)
		  ehlo_check(&buf[4]);
		if (!fmt && !esmtp_on_banner)
		  esmtp_banner_check(&buf[4]);
		if (logfp != NULL) {
		  if (debug)
		    putc('\n',logfp);
		  fprintf(logfp,
			  "%dr\t%s\n", pid, buf);
		  fflush(logfp);
		}

		if (s + 1 < cp)
		  bcopy(s+1, buf, cp-s-1);
		cp = buf + (cp-s-1);
		s = buf;
		--s;		/* incremented in for() stmt */
		/* fall through */
	      case 2:		/* saw \n, 1st char on line */
	      case 3:		/* 2nd char on line */
	      case 4:		/* 3rd char on line */
		if (i == 1
		    || (isascii(*s) && isdigit(*s)))
		  ++i;
		else
		  /* silently look for num. code lines */
		  i = 1;
		break;
	      case 5:		/* 4th char on line */
		i = (*s == '-');
		break;
	      }
	      if (debug) {
		if (i%8 == 0) {
		  if (i > 0)
		    putc('\n',logfp);
		  fprintf(logfp,
			  "%d#\t", pid);
		}
		if (isprint(*s)) 
		  putc(*s, logfp);
		else
		  fprintf(logfp,
			  "\\%03o", *s);
	      }
	    }
	    if (debug) {
	      fprintf(logfp, "\n");
	      fflush(logfp);
	    }
	  } else if (r == -1) {
#ifdef	TIMEOUT
	    if (gotalarm) {
	      sprintf(remotemsg, "466 (Timeout on %sSMTP responce read, Cmd: %s)",
		      fmt == NULL ? "initial ":"", fmt);
	      notaryreport(NULL,NULL,remotemsg);
	    } else
#endif /* TIMEOUT */
	      {
		se = strerror(errno);
		sprintf(remotemsg, "500 (Error on %sSMTP responce read: %s, Cmd: %s)",
			fmt == NULL ? "initial ":"", se, fmt);
		notaryreport(NULL,NULL,remotemsg);
	      }
	    dflag = 0;
	    firstmx = 0;
	    if (verboselog)
	      fprintf(verboselog,"%s\n",remotemsg);
	    return EX_TEMPFAIL;
	  } else {
	    sprintf(remotemsg, "500 (Server hung up on us! Cmd: %s)",fmt);
	    notaryreport(NULL,NULL,remotemsg);
	    dflag = 0;
	    if (verboselog)
	      fprintf(verboselog,"%s\n",remotemsg);
	    return EX_TEMPFAIL;
	  }
	  /* Exit if the last thing we read was a LF and we're on the
	     last line (in case of multiline response).  This
	     also takes care of the required CRLF termination */
	} while (cp < buf+sizeof buf && !(i == 0 && *(cp-1) == '\n'));

	if (cp >= buf+sizeof buf) {
	  strcpy(remotemsg,"500 (SMTP Response overran input buffer!)");
	  notaryreport(NULL,NULL,remotemsg);
	  dflag = 0;
	  if (verboselog)
	    fprintf(verboselog,"%s\n",remotemsg);
	  return EX_TEMPFAIL;
	}
	*--cp = '\0';	/* kill the LF */
	if (cp - buf < 5) {
	  sprintf(remotemsg, "500 (SMTP response '%s' unexpected!)", buf);
	  notaryreport(NULL,NULL,remotemsg);
	  dflag = 0;
	  if (verboselog)
	    fprintf(verboselog,"%s\n",remotemsg);
	  return EX_TEMPFAIL;
	}
	--cp;
	/* trim trailing whitespace */
	while (isascii(*cp) && isspace(*cp))
	  --cp;
	*++cp = '\0';
	for (i = 0; i < 4; ++i)		/* can't happen, right? wrong... */
	  if (buf[i] == ' ')
	    break;
	if (i == 4) --i;
	ch = buf[i];
	buf[i] = '\0';
	response = atoi(buf);
	if (logfp != NULL) {
	  fprintf(logfp, "%dr\t%s%c%s\n", pid, buf, ch, &buf[i+1]);
	  fflush(logfp);
	}
	buf[i] = ch;

	if (within_ehlo)
	  ehlo_check(&buf[4]);
	if (!fmt && !esmtp_on_banner)
	  esmtp_banner_check(&buf[4]);

	rmsgappend("\r->> %s",buf);
	if (saverpt)
	  notarystatsave(buf);

	
	dflag = 0;
	switch (response) {
	case 211: /* System status, or system help reply */
	case 214: /* Help message */
	case 220: /* <domain> Service ready */
	case 221: /* <domain> Service closing transmission channel */
	case 250: /* Requested mail action okay, completed */
	case 251: /* User not local; will forward to <forward-path> */
        case 255: /* Something the PMDF 4.1 returns.. for EHLO */
	case 354: /* Start mail input; end with <CRLF>.<CRLF> */
		return EX_OK;
	case 421: /* <domain> Service not available, closing transmission channel */
	case 450: /* Requested mail action not taken: mailbox unavailable */
	case 451: /* Requested action aborted: local error in processing */
	case 452: /* Requested action not taken: insufficient system storage */
		return EX_TEMPFAIL;
	case 500: /* Syntax error, command unrecognized */
	case 501: /* Syntax error in parameters or arguments */
		return EX_USAGE;
	case 502: /* Command not implemented */
		return EX_PROTOCOL;
	case 503: /* Bad sequence of commands */
		return EX_TEMPFAIL;
	case 504: /* Command parameter not implemented */
		return EX_PROTOCOL;
	case 550: /* Requested action not taken: mailbox unavailable */
	case 551: /* User not local; please try <forward-path> */
		return EX_NOUSER;
	case 552: /* Requested mail action aborted: exceeded storage allocation */
		return EX_TEMPFAIL;
	case 553: /* Requested action not taken: mailbox name not allowed */
		return EX_NOUSER;
	case 554: /* Transaction failed */
		return EX_UNAVAILABLE;
	}
	switch (response/100) {
	case 2:
	case 3:
		return EX_OK;
	case 4:
		return EX_TEMPFAIL;
	case 5:
		return EX_UNAVAILABLE;
	}
	return EX_TEMPFAIL;
}


int
smtp_ehlo(smtpfp, verboselog, fmt, arg)
	FILE *smtpfp[2], *verboselog;
	char *fmt, *arg;
{
	int rc;
	within_ehlo = 1;
	ehlo_capabilities = 0;
	rc = smtpwrite(smtpfp, verboselog, 1, fmt, arg);
	within_ehlo = 0;
	return rc;
}

/*
 * In theory, this should modify the command that ps shows for this process.
 * This is known to not be portable, hopefully it will break badly on systems
 * where it doesn't work.
 */

/* VARARGS */
void
report(va_alist)
	va_dcl
{
	va_list	ap;
	char *cp, buf[8192];
	
	va_start(ap);
	sprintf(buf, "-%s <%ld ", remotehost, atol(msgfile));
#ifdef	notdef
	if (logfp)
	  sprintf(buf+strlen(buf), ">>%s ", logfile);
	strcat(buf, "# ");
#endif
	cp = va_arg(ap, char *);
#ifdef	USE_VFPRINTF
	vsprintf(buf+strlen(buf), cp, ap);
#else	/* !USE_VFPRINTF */
	sprintf(buf+strlen(buf), cp, va_arg(ap, char *));
#endif	/* USE_VFPRINTF */
	for (cp = buf+strlen(buf); cp < buf + (eocmdline - cmdline); ++cp)
	  *cp = ' ';
	buf[eocmdline-cmdline] = '\0';
	strcpy(cmdline, buf);
	va_end(ap);
}

void
stashmyaddresses(host)
	char *host;
{
	int naddrs;
	struct hostent *hp;

	if ((hp = gethostbyname(host)) == NULL)
	  return;

	naddrs = 0;
	for (hp_init(hp); *hp_getaddr() != NULL; hp_nextaddr())
	  ++naddrs;
	nmyaddrs = naddrs;
	if ((myaddrs = (char *)malloc(nmyaddrs * hp->h_length)) == NULL)
	  return;
	for (hp_init(hp); *hp_getaddr() != NULL; hp_nextaddr())
	  bcopy(*hp_getaddr(), &myaddrs[(--naddrs)*hp->h_length], hp->h_length);
}

int
matchmyaddress(addr,len)
	char *addr;
	int len;
{
	int i;

	if (!myaddrs) return 0; /* Don't know my addresses ! */
	for (i = 0; i < nmyaddrs; ++i) {
	  /* if this is myself, skip to next MX host */
	  if (memcmp(addr, &myaddrs[i*len],len) == 0)
	    return 1;
	}
	return 0;
}

int
matchmyaddresses(hp)
	struct hostent *hp;
{
	char **alist;
#ifndef	h_addr	/* no h_addr macro -- presumably 4.2BSD or earlier.. */
	char *fakealist[2];

	alist = fakealist;
	fakealist[0] = hp->h_addr;
	fakealist[1] = NULL;
#else
	alist = hp->h_addr_list;
#endif
	while (alist && *alist) {
	  if (matchmyaddress(*alist,hp->h_length))
	    return 1;
	  ++alist;
	}
	return 0;
}

#ifdef	BIND

typedef union {
	HEADER qb1;
	char qb2[PACKETSZ];
} querybuf;

int
getmxrr(host, mx, maxmx, depth, mxcntp)
	char *host;
	struct mxdata mx[];
	int maxmx, depth;
	int *mxcntp; /* Return the NUMBER of MXes that were found,
			including those that were considered faulty.
			Will count all MXes. */
{
	HEADER *hp;
	msgdata *eom, *cp;
	struct hostent *hent;
	querybuf qbuf, answer;
	struct mxdata mxtemp;
	msgdata buf[8192], realname[8192];
	int n, i, j, nmx, ancount, qdcount, maxpref;
	u_short type;
	int saw_cname = 0;

	if (depth == 0)
	  h_errno = 0;

	if (depth > 3) {
	  sprintf(remotemsg,"500 (DNS: Recursive CNAME on '%s')",host);
	  notaryreport(NULL,NULL,remotemsg);
	  fprintf(stderr, "%s\n", remotemsg);
	  return EX_NOHOST;
	}


	n = res_mkquery(QUERY, host, C_IN, T_MX, (char *)NULL, 0, NULL,
			(char *)&qbuf, sizeof qbuf);
	if (n < 0) {
	  fprintf(stderr, "res_mkquery failed\n");
	  sprintf(remotemsg,
		  "466 (Internal: res_mkquery failed on host: %s)",host);
	  notaryreport(NULL,NULL,remotemsg);
	  return EX_SOFTWARE;
	}
	n = res_send((char *)&qbuf, n, (char *)&answer, sizeof answer);
	if (n < 0) {
	  sprintf(remotemsg,
		  "466 (Internal: res_send failed on host: %s)",host);
	  notaryreport(NULL,NULL,remotemsg);
	  return EX_TEMPFAIL;
	}
	eom = (msgdata *)&answer + n;
	/*
	 * find first satisfactory answer
	 */
	hp = (HEADER *) &answer;
	ancount = ntohs(hp->ancount);
	qdcount = ntohs(hp->qdcount);
	if (hp->rcode != NOERROR || ancount == 0) {
	  switch (hp->rcode) {
	  case NXDOMAIN:
	    /* Non-authoritative iff response from cache.
	     * Old BINDs used to return non-auth NXDOMAINs
	     * due to a bug; if that is the case by you,
	     * change to return EX_TEMPFAIL iff hp->aa == 0.
	     */
	    sprintf(remotemsg, "500 (Bind: no such domain: %s)", host);
	    notaryreport(NULL,NULL,remotemsg);
	    return EX_NOHOST;
	  case SERVFAIL:
	    sprintf(remotemsg, "500 (Bind: server failure: %s)", host);
	    notaryreport(NULL,NULL,remotemsg);
	    return EX_TEMPFAIL;
#ifdef OLDJEEVES
	    /*
	     * Jeeves (TOPS-20 server) still does not
	     * support MX records.  For the time being,
	     * we must accept FORMERRs as the same as
	     * NOERROR.
	     */
	  case FORMERR:
#endif
	  case NOERROR:
	    mx[0].host = NULL;
	    mxcount = 0;
	    return EX_OK;
#ifndef OLDJEEVES
	  case FORMERR:
#endif
	  case NOTIMP:
	  case REFUSED:
	    sprintf(remotemsg, "500 (Bind: unsupported query: %s)", host);
	    notaryreport(NULL,NULL,remotemsg);
	    return EX_NOPERM;
	  }
	  sprintf(remotemsg, "500 (Bind: unknown error, MX info unavailable: %s)", host);
	  notaryreport(NULL,NULL,remotemsg);
	  return EX_UNAVAILABLE;
	}
	nmx = 0;
	cp = (msgdata *)&answer + sizeof(HEADER);
	for (; qdcount > 0; --qdcount)
#if	defined(BIND_VER) && (BIND_VER >= 473)
		cp += dn_skipname(cp, eom) + QFIXEDSZ;
#else	/* !defined(BIND_VER) || (BIND_VER < 473) */
		cp += dn_skip(cp) + QFIXEDSZ;
#endif	/* defined(BIND_VER) && (BIND_VER >= 473) */
	realname[0] = '\0';
	maxpref = -1;
	while (--ancount >= 0 && cp < eom && nmx < maxmx-1) {
	  n = dn_expand((msgdata *)&answer, eom, cp, buf, sizeof buf);
	  if (n < 0)
	    break;
	  cp += n;
	  type = _getshort(cp);
	  cp += sizeof(u_short);
	  /*
	     class = _getshort(cp);
	     */
	  cp += sizeof(u_short) + sizeof(u_long);
	  n = _getshort(cp);
	  cp += sizeof(u_short);
	  if (type == T_CNAME) {
	    cp += dn_expand((msgdata *)&answer, eom, cp,
			    realname, sizeof realname);
	    saw_cname = 1;
	    continue;
	  } else if (type != T_MX)  {
	    cp += n;
	    continue;
	  }
	  mx[nmx].pref = _getshort(cp);
	  cp += sizeof(u_short); /* MX preference value */
	  n = dn_expand((msgdata *)&answer, eom, cp, buf, sizeof buf);
	  if (n < 0)
	    break;
	  cp += n;

	  hent = gethostbyname(buf); /* This resolves CNAME, should it
					should not in case of MX server */
	  if (!hent) {
	    *mxcntp += 1;
	    continue;		/* Well well.. spurious! */
	  }
	  if (cistrcmp(hent->h_name, myhostname) == 0)
	    maxpref = mx[nmx].pref;
	  if (matchmyaddresses(hent))
	    maxpref = mx[nmx].pref;

	  if ((mx[nmx].host = (msgdata *)emalloc(strlen(buf)+1)) == NULL) {
	    fprintf(stderr, "Out of virtual memory!\n");
	    exit(EX_OSERR);
	  }
	  strcpy(mx[nmx].host, buf);
	  /* [mea] Canonicalize this target & get A records.. */
	  ++nmx;
	}
	if (nmx == 0 && realname[0] != '\0' &&
	    cistrcmp(host,(char*)realname)) {
	  /* do it recursively for the real name */
	  return getmxrr((char *)realname, mx, maxmx, depth+1, mxcntp);
	} else if (nmx == 0) {
	  /* "give it benefit of doubt" */
	  mx[0].host = NULL;
	  mxcount = 0;
	  return EX_OK;
	}
	for (i = 0; i < nmx; ++i) {
	}

	/* discard MX RRs with a value >= that of  myhost */
	if (maxpref >= 0) {
	  for (n = i = 0; n < nmx; ++n) {
	    if (mx[n].pref >= maxpref) {
	      free(mx[n].host);
	      mx[n].host = NULL;
	      ++i;
	    }
	  }
	  if (i == nmx) {	/* we are the best MX, do it another way */
	    mx[0].host = NULL;
	    mxcount = 0;
	    return EX_OK;
	  }
	}
#ifdef	RFC974
	/* discard MX's that do not support SMTP service */
	for (n = 0; checkwks && n < nmx; ++n) {
	  if (mx[n].host == NULL)
	    continue;
	  strcpy(buf, mx[n].host);
	  /* It is an MX, it CAN'T have CNAME ! */
	  if (!getrrtype(buf, sizeof buf, T_WKS, 0)) {
	    free(mx[n].host);
	    mx[n].host = NULL;
	  }
	}
#endif	/* RFC974 */
	/* determine how many are left */
	for (i = 0, n = 0; i < nmx; ++i) {
	  if (mx[i].host == NULL)
	    continue;
	  *mxcntp += 1;
	  if (n < i) {
	    mx[n] = mx[i];
	    mx[i].host = NULL;
	  }
	  ++n;			/* found one! */
	}
	if (n == 0) {/* MX's exist, but their WKS's show no TCP smtp service */
	  sprintf(remotemsg, "500 (Bind: MX host does not support SMTP: %s)", host);
	  notaryreport(NULL,NULL,remotemsg);
	  return EX_UNAVAILABLE;
	}
	nmx = n;
	/* sort the records */
	for (i = 0; i < nmx; i++) {
	  for (j = i + 1; j < nmx; j++) {
	    if (mx[i].pref > mx[j].pref) {
	      mxtemp = mx[i];
	      mx[i] = mx[j];
	      mx[j] = mxtemp;
	    }
	  }
	}
	mx[nmx].host = NULL;
	mxcount = nmx;
	return EX_OK;
}

int
getrr(host, hbsize, rrtype, cnamelevel)	/* getrrtype with completion */
	char *host;
	int hbsize;
	u_short rrtype;
	int cnamelevel;
{
	int rval;
	char buf[BUFSIZ], **domain;

	if ((rval = getrrtype(host, hbsize, rrtype, cnamelevel)) > 0)
	  return rval;
	for (domain = _res.dnsrch; *domain != NULL; domain++) {
	  sprintf(buf, "%s.%s", host, *domain);
	  if ((rval = getrrtype(buf, BUFSIZ, rrtype, cnamelevel)) > 0) {
	    strncpy(host, buf, hbsize<BUFSIZ?hbsize:BUFSIZ);
	    host[hbsize - 1] = '\0';
	    return rval;
	  }
	}
	return 0;
}

int
getrrtype(host, hbsize, rrtype, cnamelevel)
	char *host;
	int hbsize;
	int cnamelevel;
	u_short rrtype;
{

	HEADER *hp;
	msgdata *eom, *cp;
	querybuf buf, answer;
	int n, ancount, qdcount, ok;
	u_short type;
	msgdata nbuf[BUFSIZ];
	int first;
	char *cnamestr = NULL;

	n = res_mkquery(QUERY, host, C_IN, rrtype, (char *)NULL, 0, NULL,
			(char *)&buf, sizeof(buf));
	if (n < 0) {
	  fprintf(stderr, "res_mkquery failed\n");
	  h_errno = NO_RECOVERY;
	  return -2;
	}
	n = res_send((char *)&buf, n, (char *)&answer, sizeof(answer));
	if (n < 0) {
	  h_errno = TRY_AGAIN;
	  return -1;
	}
	eom = (msgdata *)&answer + n;
	/*
	 * find first satisfactory answer
	 */
	hp = (HEADER *) &answer;
	ancount = ntohs(hp->ancount);
	qdcount = ntohs(hp->qdcount);
	h_errno = 0;
	/*
	 * We don't care about errors here, only if we got an answer
	 */
	if (ancount == 0) {
	  if (rrtype == T_CNAME && hp->rcode == NOERROR) {
	    if (qdcount > 0 && strchr(host, '.') == NULL) {
	      cp = (msgdata *)&answer + sizeof(HEADER);
	      if (dn_expand((msgdata *)&answer,
			    eom, cp, host, hbsize) >= 0) {
		if (host[0] == '\0') {
		  host[0] = '.';
		  host[1] = '\0';
		}
	      }
	    }
	    return 1;
	  }
	  return 0;
	}
	cp = (msgdata *)&answer + sizeof(HEADER);
	for (; qdcount > 0; --qdcount) {
#if	defined(BIND_VER) && (BIND_VER >= 473)
	  cp += dn_skipname(cp, eom) + QFIXEDSZ;
#else	/* !defined(BIND_VER) || (BIND_VER < 473) */
	  cp += dn_skip(cp) + QFIXEDSZ;
#endif	/* defined(BIND_VER) && (BIND_VER >= 473) */
	}
	first = 1;
	ok = rrtype != T_WKS;
	while (--ancount >= 0 && cp < eom) {
	  if ((n = dn_expand((msgdata *)&answer, eom, cp, nbuf,
			     sizeof(nbuf))) < 0)
	    break;
	  if (first) {
	    strncpy(host, nbuf, hbsize);
	    host[hbsize - 1] = '\0';
	    first = 0;
	  }
	  cp += n;
	  type = _getshort(cp);
	  cp += sizeof(u_short); /* type */
	  cp += sizeof(u_short); /* class */
	  cp += sizeof(u_long);	/* ttl */
	  n = _getshort(cp);
	  cp += sizeof(u_short); /* dlen */
	  if (type != rrtype) {
	    cp += n;
	    continue;
	  }
	  /*
	   * Assume that only one rrtype will be found.  More
	   * than one is undefined.
	   */
	  if (type == T_WKS) {
	    msgdata *nextcp = cp + n;
	    /* If we have seen a WKS, it had better have SMTP,
	     * however in absence of a WKS, assume SMTP.
	     */
	    if (n < sizeof (u_long) + 1) {
	      cp = nextcp;	/* ? */
	      continue;
	    }
	    ok = 0;
	    cp += sizeof(u_long); /* skip IP address */
	    if (*cp++ == IPPROTO_TCP) {	/* check protocol */
	      if (cp + (IPPORT_SMTP/8) < nextcp
		  && (*(cp+(IPPORT_SMTP/8))
		      & (0x80>>IPPORT_SMTP%8)))
		return 1;
	    }
	    cp = nextcp;
	    continue;
	  }
	  /* Special processing on T_CNAME ??? */
	  if ((n = dn_expand((msgdata *)&answer, eom, cp, nbuf,
			     sizeof(nbuf))) < 0)
	    break;
	  strncpy(host, nbuf, hbsize);
	  host[hbsize - 1] = '\0';
	  return ok;
	}
	return 0;
}

/*
 * This is the callback function for ctlopen.  It should return 0 to reject
 * an address, and 1 to accept it.  This routine will only be used if we've
 * been asked to check MX RR's for all hosts for applicability. Therefore we
 * check whether the addr_host has an MX RR pointing at the host that we have
 * an SMTP connection open with.  Return 1 if it is so.
 * [mea] This also understands routermxes data.
 */

int
rightmx(spec_host, addr_host)
	char	*spec_host, *addr_host;
{
	struct mxdata mxh[MAXFORWARDERS];
	int	i;
	int	mxcnt;

	if (cistrcmp(spec_host, addr_host) == 0)
	  return 1;
	if (remotehost[0] == '\0')
	  return 0;

	mxh[0].host = NULL;
	mxcnt = 0;
	switch (getmxrr(addr_host, mxh, sizeof mxh/sizeof mxh[0], 0, &mxcnt)) {
	case EX_OK:
	  if (mxh[0].host == NULL)
	    return cistrcmp(addr_host, remotehost) == 0;
	  break;
	default:
	  return 0;
	}
	for (i = 0; mxh[i].host != NULL; ++i) {
	  if (cistrcmp(mxh[i].host, remotehost) == 0) {
	    while (mxh[i].host != NULL)
	      free(mxh[i++].host);
	    return 1;
	  }
	  free(mxh[i].host);
	}
	return 0;
}
#endif	/* BIND */

/*
 * [mea] matchroutermxes()
 * like rightmx above, only a lot more light-weight...
 */
int
matchroutermxes(spec_host, ap)
	char *spec_host;
	struct address *ap;
{
	char **mxes = ap->routermxes;

	if (cistrcmp(spec_host, ap->host) == 0)
	  return 1;
	if (*remotehost == 0)
	  return 0;

	while (*mxes) {
	  if (cistrcmp(spec_host,*mxes)==0) return 1; /* Found it */
	  ++mxes;
	}
	return 0;
}


/* When data is clean 7-BIT, do:  *flag_ptr = (*flag_ptr) << 1  */
void
check_7bit_downgrade(mfd, flag_ptr)
int mfd;
int *flag_ptr;
{
	register int i;
	register int bufferfull;
	int lastwasnl;
	long mfd_pos;

	/* can we use cache of message body data ? */
	if (readalready != 0) {
	  for (i=0; i<readalready; ++i)
	    if (128 & (let_buffer[i]))
	      return;		/* Not clean ! */
	}

	/* we are assumed to be positioned properly at start of message body */
	bufferfull = 0;

	mfd_pos = tell(mfd);
	
	while ((i = read(mfd, let_buffer, sizeof let_buffer)) != 0) {
	  if (i < 0) {
	    /* ERROR ?!?!? */
	    if (bufferfull > 1) readalready = 0;
	    return;
	  }
	  lastwasnl = (let_buffer[i-1] == '\n');
	  readalready = i;
	  bufferfull++;
	  for (i=0; i<readalready; ++i)
	    if (128 & (let_buffer[i])) {
	      lseek(mfd,mfd_pos,0);
	      if (bufferfull > 1) readalready = 0;
	      return;		/* Not clean ! */
	    }
	}
	/* Got to EOF, and still it is clean 7-BIT! */
	lseek(mfd,mfd_pos,0);
	if (bufferfull > 1)	/* not all in memory, need to reread */
	  readalready = 0;

	/* Set indication! */
	*flag_ptr = (*flag_ptr) << 1;
}

void
notarystatsave(smtpline)
char *smtpline;
{
	int len = strlen(smtpline)+8;
	char *str = alloca(len);
	char *s = str;

	*s++ = *smtpline++;
	*s++ = *smtpline++;
	*s++ = *smtpline++;
	*s++ = ' ';
	if (len < 11) {
	  notaryreport(NULL,NULL,str);
	  return;
	}
	*s++ = '(';
	while (*smtpline) {
	  switch (*smtpline) {
	    case '(':
	        *s++ = '[';
		break;
	    case ')':
		*s++ = ']';
		break;
	    default:
		*s++ = *smtpline;
	  }
	  ++smtpline;
	}
	*s++ = ')';
	*s = 0;
	notaryreport(NULL,NULL,str);
}
